#!/bin/sh
#
# Copyright (c) 2004-2005 rncbc aka Rui Nuno Capela.
# All rights reserved.
#
# /etc/init.d/rtirq
#
# Startup script for realtime-preempt enabled kernels.
#
# chkconfig: 35 81 19
# description: Realtime IRQ thread tunning.
#
### BEGIN INIT INFO
# Provides:          rtirq
# Required-Start:    $syslog $local_fs
# Should-Start: $time alsa alsasound hotplug
# Required-Stop:     $syslog $local_fs
# Should-Stop: $time alsa alsasound hotplug
# Default-Start:     3 5
# Default-Stop:      0 1 2 6
# Short-Description: Realtime IRQ thread tunning.
# Description:       Change the realtime scheduling policy
#	and priority of relevant system driver IRQ handlers.
### END INIT INFO
#


# Won't work without those binaries.
for DIR in /sbin /usr/sbin /bin /usr/bin /usr/local/bin; do
	[ -z "${RTIRQ_CHRT}"  -a -x ${DIR}/chrt  ] &&  RTIRQ_CHRT=${DIR}/chrt
	[ -z "${RTIRQ_PIDOF}" -a -x ${DIR}/pidof ] && RTIRQ_PIDOF=${DIR}/pidof
done

# Check for missing binaries (stale symlinks should not happen)
[ -n "${RTIRQ_CHRT}" -a -x ${RTIRQ_CHRT} ] || {
	echo "`basename $0`: chrt: not installed."  
	[ "$1" = "stop" ] && exit 0 || exit 5
}

[ -n "${RTIRQ_PIDOF}" -a -x ${RTIRQ_PIDOF} ] || {
	echo "`basename $0`: pidof: not installed."  
	[ "$1" = "stop" ] && exit 0 || exit 5
}

# Check for existence of needed config file and read it.
RTIRQ_CONFIG=/etc/sysconfig/rtirq
[ -r ${RTIRQ_CONFIG} ] || {
	echo "`basename $0`: ${RTIRQ_CONFIG}: not found."
	[ "${RTIRQ_ACTION}" = "stop" ] && exit 0 || exit 6
}

# Read configuration.
. ${RTIRQ_CONFIG}

# Colon delimited trail list of already assigned IRQ numbers,
# preventind lower priority override due to shared IRQs.
RTIRQ_TRAIL=":"

# 
# Reset policy of all IRQ threads out there. 
#
function rtirq_reset ()
{
	# PIDS=`ps -eo pid,class | egrep '(FF|RR)' | awk '{print $1}'`
	PIDS=`ps -eo pid,comm | grep IRQ | awk '{print $1}'`
	for PID in ${PIDS}
	do
		${RTIRQ_CHRT} --pid --other 0 ${PID}
	done
}

#
# IRQ thread handler policy prioritizer, by IRQ number.
#
function rtirq_exec_num ()
{
	ACTION=$1
	NAME=$2
	PRI2=$3
	IRQ=$4
	# Check the services that are to be (un)threaded.
	if [ -n "`echo :${RTIRQ_NON_THREADED}: | sed 's/ /:/' | grep :${NAME}:`" ]
	then
		for THREADED in /proc/irq/${IRQ}/*/threaded
		do
			if [ -f "${THREADED}" ]
			then
				case ${ACTION} in
				start)
					echo 0 > "${THREADED}"
					;;
				stop)
					echo 1 > "${THREADED}"
					;;
				esac
			fi
		done
	fi
	# And now the proper threading prioritization...
	PID=`${RTIRQ_PIDOF} "IRQ ${IRQ}"`
	if [ $((${PID})) -gt 0 -a -z "`echo ${RTIRQ_TRAIL} | grep :${IRQ}:`" ]
	then 
		PREPEND="`basename $0`: ${ACTION} [${NAME}] irq=${IRQ} pid=${PID}"
		case ${ACTION} in
		start)
			PREPEND="${PREPEND} prio=${PRI2}"
			if ${RTIRQ_CHRT} --pid --fifo ${PRI2} ${PID}
			then
				echo "${PREPEND}: OK."
			else 
				echo "${PREPEND}: FAILED."
			fi
			;;
		stop)
			if ${RTIRQ_CHRT} --pid --other 0 ${PID}
			then
				echo "${PREPEND}: OK."
			else 
				echo "${PREPEND}: FAILED."
			fi
			;;
		status)
			echo "${PREPEND}: " && ${RTIRQ_CHRT} --pid --verbose ${PID}
			;;
		*)
			echo "${PREPEND}: ERROR."
			;;
		esac
		RTIRQ_TRAIL=":${IRQ}${RTIRQ_TRAIL}"
	fi
}

#
# IRQ thread handler policy prioritizer, by service name.
#
function rtirq_exec_name ()
{
	ACTION=$1
	NAME=$2
	PRI1=$3
	IRQS=`grep "${NAME}" /proc/interrupts | awk -F: '{print $1}'`
	for IRQ in ${IRQS}
	do
		rtirq_exec_num ${ACTION} ${NAME} ${PRI1} ${IRQ}
		PRI1=$((${PRI1} - 1))
	done
}


#
# Main executive.
#
function rtirq_exec ()
{
	ACTION=$1
	# Check configured base priority.
	PRI0=${RTIRQ_PRIO_HIGH}
	[ $((${PRI0})) -gt 90 ] && PRI0=90
	[ $((${PRI0})) -lt 10 ] && PRI0=10
	# Check configured priority decrease step.
	DECR=${RTIRQ_PRIO_DECR}
	[ $((${DECR})) -gt 20 ] && DECR=20
	[ $((${DECR})) -lt  1 ] && DECR=1
	# Process all configured service names...
	for NAME in ${RTIRQ_NAME_LIST}
	do
		case ${NAME} in
		snd)
			PRI1=${PRI0}
			IRQS=`grep irq /proc/asound/cards | sed 's/.* irq \(.*\)/\1/'`
			for IRQ in ${IRQS}
			do
				rtirq_exec_num ${ACTION} ${NAME} ${PRI1} ${IRQ}
				PRI1=$((${PRI1} - 1))
			done
			;;
		usb)
			rtirq_exec_name ${ACTION} "ohci_hcd" ${PRI0}
			rtirq_exec_name ${ACTION} "uhci_hcd" ${PRI0}
			rtirq_exec_name ${ACTION} "ehci_hcd" ${PRI0}
			;;
		*)
			rtirq_exec_name ${ACTION} ${NAME} ${PRI0}
			;;  
		esac
		[ ${PRI0} -gt ${DECR} ] && PRI0=$((${PRI0} - ${DECR}))
	done
}


#
# Main procedure line.
#
case $1 in
*start)
	if [ "${RTIRQ_RESET_ALL}" = "yes" -o "${RTIRQ_RESET_ALL}" = "1" ]
	then
		rtirq_reset
	fi
	rtirq_exec start
	;;
stop)
	if [ "${RTIRQ_RESET_ALL}" = "yes" -o "${RTIRQ_RESET_ALL}" = "1" ]
	then
		rtirq_reset
	#else
	#  rtirq_exec stop
	fi
	;;
reset)
	if [ "${RTIRQ_RESET_ALL}" = "yes" -o "${RTIRQ_RESET_ALL}" = "1" ]
	then
		rtirq_reset
	else
		rtirq_exec stop
	fi
	;;
status)
	echo
	#rtirq_exec status
	ps -eo pid,class,rtprio,ni,pri,pcpu,stat,comm --sort -rtprio \
		| egrep '(^[[:blank:]]*PID|IRQ)' \
		| awk 'BEGIN {
			while (getline IRQLine < "/proc/interrupts") {
				split(IRQLine, IRQSplit, ":[[:blank:]|0-9]+");
				if (match(IRQSplit[1], "^[[:blank:]]*[0-9]+$")) {
					gsub("[^[:blank:]]+PIC[^[:blank:]]*[[:blank:]]+" \
						"|[/|0-9]+$", "", IRQSplit[2]);
					IRQTable[IRQSplit[1] + 0] = IRQSplit[2];
				}
			}
		}	{ print $0"\t"IRQTable[$9]; }'
	echo
	;;
*)
	echo
	echo "  Usage: $0 {[re]start|stop|reset|status}"
	echo
	exit 1
	;;
esac

exit 0
